/****************************************************************************
** $Id: qsarray_object.cpp  beta1   edited Dec 17 18:47 $
**
** Copyright (C) 2001-2002 Trolltech AS.  All rights reserved.
**
** This file is part of the Qt Script for Applications framework (QSA).
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** Licensees holding a valid QSA Beta Evaluation Version license may use
** this file in accordance with the QSA Beta Evaluation Version License
** Agreement provided with the Software.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** See http://www.trolltech.com/pricing.html or email sales@trolltech.com for
**   information about QSA Commercial License Agreements.
** See http://www.trolltech.com/gpl/ for GPL licensing information.
**
** Contact info@trolltech.com if any conditions of this licensing are
** not clear to you.
**
*****************************************************************************/

#include "qsarray_object.h"
#include "qsoperations.h"
#include "qstypes.h"
#include "qsfunction.h"
#include "qsenv.h"

QSArrayClass::QSArrayClass( QSClass *b )
    : QSWritableClass( b )
{
    addMember( "toString", QSMember( &QSArrayClass::join ) );
    addMember( "toLocaleString", QSMember( &QSArrayClass::join ) ); // ###
    addMember( "concat", QSMember( &concat ) );
    addMember( "join", QSMember( &QSArrayClass::join ) );
    addMember( "pop", QSMember( &pop ) );
    addMember( "push", QSMember( &push ) );
    addMember( "reverse", QSMember( &reverse ) );
    addMember( "shift", QSMember( &shift ) );
    addMember( "slice", QSMember( &slice ) );
    addMember( "sort", QSMember( &sort ) );
    addMember( "splice", QSMember( &splice ) );
    addMember( "unshift", QSMember( &unshift ) );
    int lengthIndex = addVariableMember( "length", AttributeNone );
    // setLength and length functions assume that lengthIndex is null
    Q_ASSERT( !lengthIndex );
}

/*!
  \reimp
*/
QSObject QSArrayClass::cast( const QSList &args ) const
{
    return construct( args );
}


QSObject QSArrayClass::construct( const QSList &args ) const
{
    QSArray result;
    unsigned int len = 0;
    QSListIterator it = args.begin();
    // a single argument might denote the array size
    if (args.size() == 1 && it->isA(NumberType)) {
	len = it->toUInt32();
	setLength( &result, len );
    }
    else {
	// initialize array
	len = args.size();
	for (unsigned int u = 0; it != args.end(); it++, u++)
	    result.put( QSString::from( u ), *it );
    }

    // array size
    return result;
}
QSArray::QSArray()
    : QSObject( QSEnv::current()->arrayClass() )
{
    setVal( new QSInstanceData(1) );
    QSArrayClass::setLength( this, 0 );
}

void QSArrayClass::write( QSObject *objPtr, const QSMember &mem,
			  const QSObject &val ) const
{
    if ( mem.name() == "length" ) {
	uint oldLen = length( objPtr );
	uint newLen = val.toUInt32();
	// shrink array if necessary
	for ( uint u = newLen; u < oldLen; u++ ) {
	    QString pr = QString::number( u );
	    if ( objPtr->hasProperty( pr ) )
		//		deleteProperty( pr ); // ####
		objPtr->put( pr, QSUndefined() );
	}
	setLength( objPtr, newLen );
    } else {
	QSWritableClass::write( objPtr, mem, val );
	bool ok;
	uint idx = mem.name().toUInt( &ok );
	if ( ok ) {
	    uint len = length( objPtr );
	    if ( idx >= len )
		setLength( objPtr, idx + 1 );
	}
    }
}

/*!
  \reimp

  Equivalent to join() with a "," seperator.
*/

QString QSArrayClass::toString( const QSObject *obj ) const
{
    return joinInternal( obj, "," );
}

/*!
  Returns the length of array object \a a.
 */

uint QSArrayClass::length( const QSObject *a )
{
    Q_ASSERT( a && a->objectType()==QSEnv::current()->arrayClass() );
    return (uint)( (QSInstanceData*) a->shVal() )->value(0)->toNumber();
}

/*!
  Sets the length property of array object \a a to \a l.
 */

void QSArrayClass::setLength( QSObject *a, uint l )
{

    ( (QSInstanceData*) a->shVal() )->setValue( 0, QSNumber(l) );
}


/*!
  Returns a string concatenating all members of this array seperated
  by \a sep.
*/

QString QSArrayClass::joinInternal( const QSObject *obj, const QString &sep )
{
    uint len = length( obj );
    QString str = "";
    for ( uint i = 0; i < len; i++ ) {
	if ( i >= 1 )
	    str += sep;
	QSObject o = obj->get( QSString::from( i ) );
	if ( !o.isA(UndefinedType) && !o.isA(NullType) )
	    str += o.toString();
    }
    return str;
}

QSObject QSArrayClass::toPrimitive( const QSObject *obj,  Type ) const
{
    return QSString( toString( obj ) );
}

QSObject QSArrayClass::concat( QSObject *o, const QSList &args )
{
    QSArray result;
    int n = 0, len = 0;
    QSObject obj = *o;
    QSListIterator it = args.begin();
    for ( ;; ) {
	if ( obj.isA( "Array" ) ) {
	    unsigned int k = 0;
	    if ( n == 0 )
		len = length( &obj );
	    while ( (int) k < len ) {
		QString p = QSString::from( k );
		if ( obj.hasProperty( p ) )
		    result.put( QSString::from( n ), obj.get( p ) );
		n++;
		k++;
	    }
	} else {
	    result.put( QSString::from( n ), obj );
	    n++;
	}
	if ( it == args.end() )
	    break;
	obj = it++;
    }
    result.put( "length", QSNumber( n ), DontEnum | DontDelete );
    return result;
}

/*!
  Returns a string contain all elements of array \a o joined by
  the first element in \a args or ',' if unspecified.
 */

QSObject QSArrayClass::join( QSObject *o, const QSList &args )
{
    QString separator = args[0].isDefined() ?
			args[0].toString() : QString( "," );
    return QSString( QSArrayClass::joinInternal( o, separator ) );
}

/*!
  Pop the last element of array \a o and return it. Returns undefined
  in case the array is empty.
*/

QSObject QSArrayClass::pop( QSObject *o, const QSList & )
{
    uint len = length( o );
    if ( len == 0 ) {
	return QSUndefined();
    } else {
	QString str = QSString::from( len - 1 );
	QSObject result = o->get( str );
	o->deleteProperty( str );
        setLength( o, len - 1 );
	return result;
    }
}

/*!
  Push all \a args to the end of array \a o.
 */

QSObject QSArrayClass::push( QSObject *o, const QSList &args )
{
    uint len = length( o );
    for ( int n = 0; n < args.size(); n++ )
	o->put( QSString::from( len + n ), args[n] );
    len += args.size();
    setLength( o, len );
    return QSNumber( len );
}

/*!
  Reverse the order of elements in array \a o and return the array.
*/

QSObject QSArrayClass::reverse( QSObject *o, const QSList & )
{
    uint len = length( o );
    uint middle = len / 2;
    QSObject thisObj = *o;
    for ( uint k = 0; k < middle; k++ ) {
	QString str = QString::number( k );
	QString str2 = QString::number( len - k - 1 );
	QSObject obj = thisObj.get( str );
	QSObject obj2 = thisObj.get( str2 );
	if ( thisObj.hasProperty( str2 ) ) {
	    if ( thisObj.hasProperty( str ) ) {
		thisObj.put( str, obj2 );
		thisObj.put( str2, obj );
	    } else {
		thisObj.put( str, obj2 );
		thisObj.deleteProperty( str2 );
	    }
	} else {
	    if ( thisObj.hasProperty( str ) ) {
		thisObj.deleteProperty( str );
		thisObj.put( str2, obj );
	    } else {
		// why delete something that's not there ? Strange.
		thisObj.deleteProperty( str );
		thisObj.deleteProperty( str2 );
	    }
	}
    }
    return thisObj;
}

/*!
  Removes the first element from array \a o and returns it.
*/

QSObject QSArrayClass::shift( QSObject *o, const QSList & )
{
    uint len = length( o );
    if ( len == 0 )
	return QSUndefined();

    QSObject result = o->get( "0" );
    for( uint k = 1; k < len; k++ ) {
	QString str = QString::number( k );
	QString str2 = QString::number( k-1 );
	if ( o->hasProperty( str ) )
	    o->put( str2, o->get( str ) );
	else
	    o->deleteProperty( str2 );
    }
    o->deleteProperty( QString::number( len - 1 ) );
    setLength( o, len - 1 );
    return result;
}

/*!
  Returns the elements \a args[0] up to (but excluding) \a args[1]
  from array \a o.
  ### meaning of negative indices
 */

QSObject QSArrayClass::slice( QSObject *o, const QSList &args )
{
    QSArray result;
    uint len = length( o );
    int begin = args[0].toUInt32();
    int end = len;
    if ( !args[1].isA( UndefinedType ) ) {
	end = args[1].toUInt32();
	if ( end < 0 )
            end += len;
        // safety tests
        if ( begin < 0 || end < 0 || begin >= end ) {
            result.put( "length", QSNumber( 0 ), DontEnum | DontDelete );
	    return result;
        }
        for( uint k = 0; k < (uint) end-begin; k++ ) {
            QString str = QString::number( k+begin );
            QString str2 = QString::number( k );
            if ( o->hasProperty( str ) )
                result.put( str2, o->get( str ) );
        }
        result.put( "length", end - begin, DontEnum|DontDelete );
    }
    return result;
}

/*!
  Sort the array \a o. If an argument is given it is used as a
  compare function.
*/

QSObject QSArrayClass::sort( QSObject *o, const QSList &args )
{
    QSObject sortFunction;
    bool useSortFunction = args[0].isDefined();
    if (useSortFunction) {
	sortFunction = args[0];
	if ( !sortFunction.isExecutable() )
	    useSortFunction = false;
    }

    uint len = length( o );
    if ( len == 0 )
	return QSUndefined();

    // "Min" sort. Not the fastest, but definitely less code than heapsort
    // or quicksort, and much less swapping than bubblesort/insertionsort.
    for ( uint i = 0; i < len-1 ; ++i ) {
	QSObject iObj = o->get( QString::number( i ) );
	uint themin = i;
	QSObject minObj = iObj;
	for ( unsigned int j = i+1; j < len; ++j ) {
	    QSObject jObj = o->get( QString::number( j ) );
	    int cmp;
	    if ( useSortFunction ) {
		QSList l;
		l.append( jObj );
		l.append( minObj );
		cmp = sortFunction.execute( l ).toInt32();
	    } else {
		cmp = ( jObj.toString() < minObj.toString() ) ? -1 : 1;
	    }
	    if ( cmp < 0 ) {
		themin = j;
		minObj = jObj;
	    }
            // Swap themin and i
            if ( themin > i ) {
                o->put( QString::number( i ), minObj );
                o->put( QString::number( themin ), iObj );
            }
        }
    }
    return *o;
}

/*! Array.splice(start, deleteCount [,item1 [,item2 [, ...]]])
  \a deleteCount elements of array \a o at index \a start are
  replaced by the following arguments.
 */

QSObject QSArrayClass::splice( QSObject *o, const QSList &args )
{
    QSArray result;

    int k;
    int len = length( o );
    int start = args[0].toInteger();
    int del   = args[1].toInteger();
    start = ( start < 0 ) ? QMAX( len+start, 0 ) : QMIN( len, start );
    del = QMIN( QMAX( del, 0 ), len-start );
    for ( k=0; k<del; k++ ) {
	QString index = QString::number( k+start );
	if ( o->hasProperty( index ) )
	    result.put( QString::number(k), o->get( index ) );
    }
    int add   = args.size()-2;
    if ( add>del ) {
	for ( k=len-del; k>start; k-- ) {
	    QString s39 = QString::number( k+del-1 );
	    QString s40 = QString::number( k+add-1 );
	    if ( o->hasProperty( s39 ) )  // 42
		o->put( s40, o->get( s39 ) );
	    else  // 45
		o->deleteProperty( s40 );
	}
    } else {
	for ( k=start; k<len-del; k++ ) {
	    QString s22 = QString::number( k+del );
	    QString s23 = QString::number( k+add );
	    if ( o->hasProperty( s22 ) )  // 25
		o->put( s23, o->get( s22 ) );
	    else // 28
		o->deleteProperty( s23 );
	} // 31
	for ( k=len; k>len-del+add; k-- )
	    o->deleteProperty( QString::number( k-1 ) );
    }
    for( k=start; k<start+add; k++ )
	o->put( QString::number(k), args[k-start+2] );

    return result;
}

/*!
  Prepend each element to the array \a o preserving order.
 */

QSObject QSArrayClass::unshift( QSObject *o, const QSList &args )
{
    uint len = length( o );
    int offset = args.size();
    for ( uint k = len+offset-1; (int)k >= offset; k-- ) {
	QString str = QString::number( k );
	QString str2 = QString::number( k-offset );
	if ( o->hasProperty( str2 ) )
	    o->put( str, o->get( str2 ) );
 	else
	    o->deleteProperty( str );
    }
    for ( int n = 0; n < offset; n++ )
	o->put( QSString::from( n ), args[n] );
    len += offset;
    setLength( o, len );
    return QSNumber( len );
}
