/****************************************************************************
** $Id: qsdate_object.cpp  beta1   edited Dec 17 19:43 $
**
** Copyright (C) 2001-2002 Trolltech AS.  All rights reserved.
**
** This file is part of the Qt Script for Applications framework (QSA).
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** Licensees holding a valid QSA Beta Evaluation Version license may use
** this file in accordance with the QSA Beta Evaluation Version License
** Agreement provided with the Software.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** See http://www.trolltech.com/pricing.html or email sales@trolltech.com for
**   information about QSA Commercial License Agreements.
** See http://www.trolltech.com/gpl/ for GPL licensing information.
**
** Contact info@trolltech.com if any conditions of this licensing are
** not clear to you.
**
*****************************************************************************/

#include "qsdate_object.h"

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include "qsoperations.h"

#ifdef Q_OS_UNIX
#  include <unistd.h>
#  include <sys/time.h>
#endif

#include <math.h>
#include <string.h>
#include <stdio.h>
#include <locale.h>
#include <time.h>
#ifdef HAVE_SYS_TIME_H
#  include <unistd.h>
#  include <sys/time.h>
#endif
#ifdef HAVE_SYS_TIMEB_H
#  include <sys/timeb.h>
#endif
#ifdef HAVE_SYS_PARAM_H
#  include <sys/param.h>
#endif

enum { ToString, ToDateString, ToTimeString, ToLocaleString,
       ToLocaleDateString, ToLocaleTimeString, ValueOf, GetTime,
       GetFullYear, GetMonth, GetDate, GetDay, GetHours, GetMinutes,
       GetSeconds, GetMilliSeconds, GetTimezoneOffset, SetTime,
       SetMilliSeconds, SetSeconds, SetMinutes, SetHours, SetDate,
       SetMonth, SetFullYear, ToUTCString,
       // non-normative properties (Appendix B)
       GetYear, SetYear, ToGMTString,
       UTCOffset = 100 };

// helper functions
QSObject parseDate(const QSString &s);
QSObject timeClip(const QSObject &t);

QSObject parseDate(const QSString &s)
{
    QString u = s.value();
    int firstSlash = u.find('/');
    if ( firstSlash == -1 )
	{
	    /* TODO parse dates like "December 25, 1995 23:15:00"*/
	    fprintf(stderr,"KJS::parseDate parsing for this format isn't implemented\n%s", u.ascii());
	    return QSNumber(0);
	}
    else
	{
	    // Found 12/31/2099 on some website -> obviously MM/DD/YYYY
	    int month = QSString::toULong( u.mid( 0,firstSlash ) );
	    int secondSlash = u.find('/',firstSlash+1);
	    //fprintf(stdout,"KJS::parseDate firstSlash=%d, secondSlash=%d\n", firstSlash, secondSlash);
	    if ( secondSlash == -1 )
		{
		    fprintf(stderr,"KJS::parseDate parsing for this format isn't implemented\n%s", u.ascii());
		    return QSNumber(0);
		}
	    int day = QSString::toULong( u.mid( firstSlash+1,
						secondSlash-firstSlash-1) );
	    int year = QSString::toULong( u.mid( secondSlash+1 ) );
	    //fprintf(stdout,"KJS::parseDate day=%d, month=%d, year=%d\n", day, month, year);
	    struct tm t;
	    memset( &t, 0, sizeof(t) );
	    year = (year > 2037) ? 2037 : year; // mktime is limited to 2037 !!!
	    t.tm_year = (year >= 0 && year <= 99) ? year : year - 1900;
	    t.tm_mon = month-1; // mktime wants 0-11 for some reason
	    t.tm_mday = day;
	    time_t seconds = mktime(&t);
	    if ( seconds == -1 )
		{
		    fprintf(stderr,"KJS::parseDate mktime returned -1.\n%s", u.ascii());
		    return QSUndefined();
		}
	    else
		return QSNumber( seconds * 1000.0 );
	}
}

QSObject timeClip(const QSObject &t)
{
    /* TODO */
    return t;
}

QSDateClass::QSDateClass( QSClass *b )
    : QSClass( b, AttributeFinal )
{
    addMember( "parse", QSMember( &parse, AttributeStatic ) );
    addMember( "utc", QSMember( &utc, AttributeStatic ) );
}

QSObject QSDateClass::construct( const QSList &args ) const
{
    double value;

    int numArgs = args.size();

    if (numArgs == 0) { // new Date() ECMA 15.9.3.3
#ifdef HAVE_SYS_TIMEB_H
#  if defined(__BORLANDC__)
	struct timeb timebuffer;
	ftime(&timebuffer);
#  else
	struct _timeb timebuffer;
	_ftime(&timebuffer);
#  endif
	double utc = floor((double)timebuffer.time * 1000.0 + (double)timebuffer.millitm);
#else
	struct timeval tv;
	gettimeofday(&tv, 0L);
	double utc = floor((double)tv.tv_sec * 1000.0 + (double)tv.tv_usec / 1000.0);
#endif // HAVE_SYS_TIMEB_H
	value = utc;
    } else if (numArgs == 1) {
	QSObject p = args[0].toPrimitive();
	if (p.isA(StringType))
	    value = parseDate( p.toString() ).toNumber();
	else
	    value = p.toNumber();
    } else {
	struct tm t;
	memset(&t, 0, sizeof(t));
	QSNumber y( args[0].toNumber() );
	/* TODO: check for NaN */
	int year = y.toInt32();
	t.tm_year = (year >= 0 && year <= 99) ? year : year - 1900;
	t.tm_mon = args[1].toInt32();
	t.tm_mday = (numArgs >= 3) ? args[2].toInt32() : 1;
	t.tm_hour = (numArgs >= 4) ? args[3].toInt32() : 0;
	t.tm_min = (numArgs >= 5) ? args[4].toInt32() : 0;
	t.tm_sec = (numArgs >= 6) ? args[5].toInt32() : 0;
	t.tm_isdst = -1;
	int ms = (numArgs >= 7) ? args[6].toInt32() : 0;
	value = mktime(&t) * 1000.0 + ms;
    }

    QSObject date( this );
    date.setVal( value );
    return date;
}

QSObject QSDateClass::cast( const QSList & ) const
{
    time_t t = time( 0L );
    QString s( ctime( &t ) );

    // return formatted string minus trailing \n
    return QSString( s.mid(0, s.length() - 1 ) );
}

bool QSDateClass::member( const QSObject *o, const QString &p,
			  QSMember *m ) const
{
    int id;
    if ( p == "toString" || p == "toUTCString" )
	id = ToString;
    else if ( p == "toDateString" )
	id = ToDateString;
    else if ( p == "toTimeString" )
	id = ToTimeString;
    else if ( p == "toLocaleString" )
	id = ToLocaleString;
    else if ( p == "toLocaleDateString" )
	id = ToLocaleDateString;
    else if ( p == "toLocaleTimeString" )
	id = ToLocaleTimeString;
    else if ( p == "valueOf" )
	id = ValueOf;
    else if ( p == "getTime" )
	id = GetTime;
    else if ( p == "getFullYear" || p == "getUTCFullYear" )
	id = GetFullYear;
    else if ( p == "toGMTString" )
	id = ToGMTString;
    else if ( p == "getMonth" || p == "getUTCMonth" )
	id = GetMonth;
    else if ( p == "getDate" || p == "getUTCDate" )
	id = GetDate;
    else if ( p == "getDay" || p == "getUTCDay" )
	id = GetDay;
    else if ( p == "getHours" || p == "getUTCHours" )
	id = GetHours;
    else if ( p == "getMinutes" || p == "getUTCMinutes" )
	id = GetMinutes;
    else if ( p == "getSeconds" || p == "getUTCSeconds" )
	id = GetSeconds;
    else if ( p == "getMilliseconds" || p == "getUTCMilliseconds" )
	id = GetMilliSeconds;
    else if ( p == "getTimezoneOffset" )
	id = GetTimezoneOffset;
    else if ( p == "setTime" )
	id = SetTime;
    else if ( p == "setMilliseconds" || p == "setUTCMilliseconds" )
	id = SetMilliSeconds;
    else if ( p == "setSeconds" || p == "setUTCSeconds" )
	id = SetSeconds;
    else if ( p == "setMinutes" || p == "setUTCMinutes" )
	id = SetMinutes;
    else if ( p == "setHours" || p == "setUTCHours" )
	id = SetHours;
    else if ( p == "setDate" || p == "setUTCDate" )
	id = SetDate;
    else if (p == "setMonth" || p == "setUTCMonth")
	id = SetMonth;
    else if ( p == "setFullYear" || p == "setUTCFullYear" )
	id = SetFullYear;
    else if ( p == "setYear" )
	id = SetYear;
    // non-normative
    else if ( p == "getYear" )
	id = GetYear;
    else if ( p == "toGMTString" )
	id = ToGMTString;
    else
	return QSClass::member( o, p, m );

    bool utc = p.find( "UTC" ) >= 0;
    m->setType( QSMember::Custom );
    m->setOwner( this );
    m->setIndex( id + utc * UTCOffset );
    m->setExecutable( TRUE );
    m->setName( p );

    return TRUE;
}

bool QSDateClass::toBoolean( const QSObject *obj ) const
{
    return bool(obj->dVal());
}

double QSDateClass::toNumber( const QSObject *obj ) const
{
    return obj->dVal();
}

QString QSDateClass::toString( const QSObject *o ) const
{
    double milli = o->dVal();
    time_t tv = (time_t) floor(milli / 1000.0);
    // struct tm *t = localtime(&tv);
    QString s = ctime( &tv );
    return s.mid( 0, s.length() - 1 );
}

// Date.parse()
QSObject QSDateClass::parse( const QSList &args )
{
    if ( args[0].isA( StringType ) )
	return parseDate( args[0].toString() );
    else
	return QSUndefined();
}

// Date.UTC()
QSObject QSDateClass::utc( const QSList &args )
{
    struct tm t;
    memset(&t, 0, sizeof(t));
    int n = args.size();
    /* TODO: check for NaN */
    int year = args[0].toInt32();
    t.tm_year = (year >= 0 && year <= 99) ? year : year - 1900;
    t.tm_mon = args[1].toInt32();
    t.tm_mday = (n >= 3) ? args[2].toInt32() : 1;
    t.tm_hour = (n >= 4) ? args[3].toInt32() : 0;
    t.tm_min = (n >= 5) ? args[4].toInt32() : 0;
    t.tm_sec = (n >= 6) ? args[5].toInt32() : 0;
    int ms = (n >= 7) ? args[6].toInt32() : 0;

    return QSNumber( mktime(&t) * 1000.0 + ms );
}

QSObject QSDateClass::invoke( QSObject *objPtr, const QSMember &mem,
			      const QSList &args ) const
{
    if ( mem.type() != QSMember::Custom )
	return QSClass::invoke( objPtr, mem, args );

    int id = mem.idx;
    bool utc = FALSE;
    if ( id >= UTCOffset ) {
	utc = TRUE;
	id -= UTCOffset;
    }

    QSObject result;
    QString s;
    const int bufsize=100;
    char timebuffer[bufsize];
    char *oldlocale = setlocale(LC_TIME,NULL);
    if (!oldlocale)
	oldlocale = setlocale(LC_ALL, NULL);
    QSObject thisObj = *objPtr;
    double milli = thisObj.toNumber();
    time_t tv = (time_t) floor(milli / 1000.0);
    int ms = int(milli - tv * 1000.0);

    struct tm *t;
    if (utc)
	t = gmtime(&tv);
    else
	t = localtime(&tv);

    switch (id) {
    case ToString:
	s = ctime(&tv);
	result = QSString(s.mid(0, s.length() - 1));
	break;
    case ToDateString:
    case ToTimeString:
    case ToGMTString:
	setlocale( LC_TIME, "C" );
	if ( id == ToDateString ) {
	    strftime(timebuffer, bufsize, "%x",t);
	} else if ( id == ToTimeString ) {
	    strftime(timebuffer, bufsize, "%X",t);
	} else {
	    t = gmtime(&tv);
	    strftime(timebuffer, bufsize, "%a, %d-%b-%y %H:%M:%S %Z", t);
	}
	setlocale(LC_TIME,oldlocale);
	result = QSString(timebuffer);
	break;
    case ToLocaleString:
	strftime(timebuffer, bufsize, "%c", t);
	result = QSString(timebuffer);
	break;
    case ToLocaleDateString:
	strftime(timebuffer, bufsize, "%x", t);
	result = QSString(timebuffer);
	break;
    case ToLocaleTimeString:
	strftime(timebuffer, bufsize, "%X", t);
	result = QSString(timebuffer);
	break;
    case ValueOf:
    case GetTime:
	result = QSNumber( milli );
	break;
    case GetYear:
	result = QSNumber( t->tm_year );
	break;
    case GetFullYear:
	result = QSNumber( 1900 + t->tm_year );
	break;
    case GetMonth:
	result = QSNumber( t->tm_mon );
	break;
    case GetDate:
	result = QSNumber( t->tm_mday );
	break;
    case GetDay:
	result = QSNumber( t->tm_wday );
	break;
    case GetHours:
	result = QSNumber( t->tm_hour );
	break;
    case GetMinutes:
	result = QSNumber( t->tm_min );
	break;
    case GetSeconds:
	result = QSNumber( t->tm_sec );
	break;
    case GetMilliSeconds:
	result = QSUndefined();
	break;
    case GetTimezoneOffset:
#if defined BSD || defined(__APPLE__)
	result = QSNumber(-( t->tm_gmtoff / 60 ) + ( t->tm_isdst ? 60 : 0 ));
#else
#  if defined(__BORLANDC__)
// ### verify daylight offset
	result = QSNumber(_timezone / 60 - (_daylight ? 60 : 0));
#  else
	result = QSNumber(( timezone / 60 - ( daylight ? 60 : 0 )));
#  endif
#endif
	break;
    case SetTime:
	milli = args[0].round();
	result = QSNumber( milli );
	//###    thisObj.setInternalValue(result);
	break;
    case SetMilliSeconds:
	ms = args[0].toInt32();
	break;
    case SetSeconds:
	t->tm_sec = args[0].toInt32();
	break;
    case SetMinutes:
	t->tm_min = args[0].toInt32();
	break;
    case SetHours:
	t->tm_hour = args[0].toInt32();
	break;
    case SetDate:
	t->tm_mday = args[0].toInt32();
	break;
    case SetMonth:
	t->tm_mon = args[0].toInt32();
	break;
    case SetFullYear:
	t->tm_year = args[0].toInt32() - 1900;
	break;
    case SetYear:
	t->tm_year = args[0].toInt32() >= 1900 ? args[0].toInt32() - 1900 : args[0].toInt32();
	break;
    }

    if (id == SetYear || id == SetMilliSeconds || id == SetSeconds ||
	id == SetMinutes || id == SetHours || id == SetDate ||
	id == SetMonth || id == SetFullYear ) {
	result = QSNumber( mktime(t) * 1000.0 + ms );
	// ### thisObj.setInternalValue(result);
    }

    return result;
}
