/****************************************************************************
** $Id: qsa.h  beta1   edited Dec 17 00:50 $
**
** Definition of the QtApplicationScript class.
**
** Copyright (C) 2001-2002 Trolltech AS.  All rights reserved.
**
** This file is part of the Qt Script for Applications framework (QSA).
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** Licensees holding a valid QSA Beta Evaluation Version license may use
** this file in accordance with the QSA Beta Evaluation Version License
** Agreement provided with the Software.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** See http://www.trolltech.com/pricing.html or email sales@trolltech.com for
**   information about QSA Commercial License Agreements.
** See http://www.trolltech.com/gpl/ for GPL licensing information.
**
** Contact info@trolltech.com if any conditions of this licensing are
** not clear to you.
**
*****************************************************************************/
#ifndef QSA_H
#define QSA_H

//
//  W A R N I N G
//  -------------
//
// This file is not part of the QSA API.  It exists for the convenience
// of other QSA sources files.  This header file will change from
// version to version without notice, and most likely be removed.
//

#include <qstring.h>
#include <qobjectlist.h>
#include <qwidgetlist.h>
#include <qvariant.h>
#include <qdatetime.h>
#include <qvaluelist.h>
#include <qsaglobal.h>
#include <qptrlist.h>

class QObject;
class QWidget;
class MainWindow;
class Project;
class QSADesignerInterface;
class QSEditor;
class FormWindow;
class QSProject;
class QSInterpreter;

struct QSA_EXPORT QSSignalHandler
{
    QObject *sender, *receiver;
    QString signal, function;
    bool operator==( const QSSignalHandler &sh ) const {
	return ( sh.sender == sender &&
		 sh.receiver == receiver &&
		 sh.signal == signal &&
		 sh.function == function );
    }
};

class QSA_EXPORT QtApplicationScript : public QObject
{
    Q_OBJECT
    Q_ENUMS( ErrorMode )
    Q_PROPERTY( ErrorMode errorMode READ errorMode WRITE setErrorMode )

    friend void deleteSelfPtr();
    friend class QSADesignerInterface;
    friend class QSEditor;
    friend class QSInterpreter;
    friend class QSInterfaceFactory;

public:
    enum ErrorMode {
	Notify,
	AskForDebug,
	Debug,
	Nothing
    };

    enum DeveloperFlags {
	GUIBuilder = 1,
	Debugger = 2,
	SignalHandlerTab = 4,
	EditableEditor = 8,
	FullDeveloper = GUIBuilder | Debugger | SignalHandlerTab | EditableEditor,
	OnlyEditor = EditableEditor | SignalHandlerTab,
	OnlySimpleEditor = EditableEditor,
	OnlyEditorAndDebugger = EditableEditor | Debugger | SignalHandlerTab
    };

    QtApplicationScript();
    ~QtApplicationScript();

    static QtApplicationScript *self();

    // now QSInterpreter
    QVariant evaluate( const QString &code, QObject *context = 0 );
    QVariant call( const QString &function, const QValueList<QVariant> &arguments, QObject *context = 0 );
    bool checkSyntax( const QString &code );

    void setErrorMode( ErrorMode m ) { em = m; }
    ErrorMode errorMode() const { return em; }

    QObject *currentContext() const;

    QStringList functionsOf( const QString &context, bool includeSignature = FALSE ) const;
    QStringList functionsOf( QObject *context, bool includeSignature = FALSE ) const;
    QStringList globalFunctions( bool includeSignature = FALSE ) const;
    QStringList classes() const;
    QStringList classesOf( const QString &context ) const;
    QStringList classesOf( QObject *context ) const;
    QStringList globalClasses() const;
    QStringList variablesOf( const QString &context ) const;
    QStringList variablesOf( QObject *context ) const;
    QStringList globalVariables() const;
    const QObjectList *objects() const;

    QSProject *currentProject() const;
    QSInterpreter *interpreter() const;

    // now QSProject
    QStringList forms() const;
    QStringList scriptFiles() const;
    QString formSource( const QString &formName ) const;
    QString formXML( const QString &formName ) const;
    QString source( const QString &fileName ) const;
    QByteArray projectData() const;

    bool eventFilter( QObject *, QEvent * );

    // now QSInterpreter
    QString objectSource( QObject *object ) const;
    void addSignalHandler( QObject *sender, const char *signal,
			   QObject *receiver, const char *qtscriptFunction );
    void addSignalHandler( QObject *sender, const char *signal, const char *qtscriptFunction );
    void removeSignalHandler( QObject *sender, const char *signal,
			      QObject *receiver, const char *qtscriptFunction );
    void removeSignalHandler( QObject *sender, const char *signal, const char *qtscriptFunction );

    // ###
    virtual void registerEditor( QSEditor *editor );
    virtual void unregisterEditor( QSEditor *editor );
    // now QSEditor
    const QPtrList<QSEditor> *editors() const;
    QSEditor *activeEditor() const;

public slots:
    // now QSProject
    void openDeveloper( int flags = FullDeveloper );
    void closeDeveloper();
    bool addFunction( const QString &functionName, const QString &functionBody = QString::null,
		      bool openDeveloper = TRUE );

    // now QSInterpreter
    void addObject( QObject *o );
    void removeObject( QObject *o );
    void clearObjects();
    void setObjectSource( QObject *object, const QString &code, bool append = FALSE );

    // old API
    void addSource( const QString &code, const QString &fileName, bool append = FALSE );
    void removeSource( const QString &fileName );
    void addForm( const QString &fileName );
    void removeForm( const QString &fileName );
    void open( const QString &scriptProject );
    void open( const QByteArray &projectData, const QString &projectName );
    // end of old API

    // now QSInterpreter
    void clear();

    // now QSProject
    void reset();

signals:
    // now QSInterpreter
    void error( const QString &message, const QString &filename, int lineNumber );

private slots:
    void objectDestroyed( QObject *o );
    void runtimeError( const QString &message, const QString &filename, int lineNumber );
    void runtimeError();
    void runFinished();
    bool checkProject();
    void formAdded( FormWindow *fw );

private:
    void stopProject( bool force = FALSE );
    void setupPluginManagers();
    void setupDesigner();
    void readdObjects();
    void readdEventHandlers();
    void checkAndStopProject() const;

private:
    MainWindow* designer;
    QObjectList appObjs;
    bool projectOpened;
    QString projectName;
    QDateTime projectLastModified;
    bool scriptRunning;
    bool projectRunning;
    ErrorMode em;
    static QtApplicationScript *selfPtr;
    bool blockErrors;
    Project *project;
    QValueList<QSSignalHandler> *signalHandlers;
    QPtrList<QSEditor> *listEditors;
    bool hadError;
    bool needRerun;
    QSProject *prj;
    QSInterpreter *qsip;
};

#endif
