/****************************************************************************
** $Id: qsproject.cpp  beta1   edited Dec 17 00:50 $
**
** Definition of the QSProject class.
**
** Copyright (C) 2001-2002 Trolltech AS.  All rights reserved.
**
** This file is part of the Qt Script for Applications framework (QSA).
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** Licensees holding a valid QSA Beta Evaluation Version license may use
** this file in accordance with the QSA Beta Evaluation Version License
** Agreement provided with the Software.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** See http://www.trolltech.com/pricing.html or email sales@trolltech.com for
**   information about QSA Commercial License Agreements.
** See http://www.trolltech.com/gpl/ for GPL licensing information.
**
** Contact info@trolltech.com if any conditions of this licensing are
** not clear to you.
**
*****************************************************************************/

#include "qsproject.h"
#include "qsa.h"

class QSProjectPrivate {
public:
};

/*!
  \class QSProject qsproject.h

  \brief The QSProject class provides control over a script project
  for the Qt Application Script interpreter.

  Qt Script for Applications (QSA) collects sets of functions and
  classes together in \e projects. Use open() to access a project,
  either from a file or from a byte array. If a file is used, QSA will
  automatically take care of saving user changes. The project's data
  can be retrieved into a byte array using projectData(). One project
  is current at any one time; close the current project with clear().

  The contents of the current project are available using forms() and
  classes() which return a list of all the forms and classes
  respectively.

  A new global function can be added with addFunction(); this function
  can also invoke \e{QSA Developer} if required. \e{QSA Developer} is
  invoked with openDeveloper() and closed with closeDeveloper().

  If \e{QSA Developer} is too complex for your application (even with
  the GUI builder functionality and/or the debugger switched off), you
  can use QSEditor to create a source code editor for Qt Script (with
  syntax highlighting, completion, auto indentation, etc), which can
  be embedded into your application user interface.

  See the \link qsa.book Manual\endlink for more explanation and
  examples.

*/

/*! \internal */

QSProject::QSProject( QtApplicationScript *s )
{
    qas = s;
    d = new QSProjectPrivate;
}

/*! \internal
 Destructor */

QSProject::~QSProject()
{
    delete d;
}

/*!  Returns all names of the forms of the project.

  \sa formSource(), formXML()
*/

QStringList QSProject::forms() const
{
    return qas->forms();
}

/*! Returns a list for all files in this project */

QStringList QSProject::scriptFiles() const
{
    return qas->scriptFiles();
}

/*! Returns the XML GUI description of the form called \a formName

  \sa forms(), formSource()
*/

QString QSProject::formXML( const QString &formName ) const
{
    return qas->formXML( formName );
}

/*! Returns the code for the form called \a formName

  \sa forms(), formXML()
*/

QString QSProject::formSource( const QString &formName ) const
{
    return qas->formSource( formName );
}

/*! Returns the source of the file \a fileName in this project
 */

QString QSProject::source( const QString &fileName ) const
{
    return qas->source( fileName );
}

/*!
  Returns the whole project as a data block (QByteArray).

  This is useful for keeping the project data with a document. For
  example, you could take the data returned by this function and store
  it in a document's file. Later, when the document was opened, you
  could extract this data from the document and pass it to the
  overload of open() that takes a QByteArray to reconstruct the
  project.
*/

QByteArray QSProject::projectData() const
{
    return qas->projectData();
}

/*! Adds the code \a code under the name \a fileName to the project as
  global code

  If there has been added code already under the name \a fileName, \a
  code is appended to this code if \a append is TRUE, otherwise the
  currently set code is replaced with \a code.

  \sa removeSource()
*/

void QSProject::addSource( const QString &code, const QString &fileName, bool append )
{
    qas->addSource( code, fileName, append );
}

/*! Removes the code which was added under the filename \a fileName.

  \sa addSource()
*/

void QSProject::removeSource( const QString &fileName )
{
    qas->removeSource( fileName );
}

/*! Loads the form \a fileName and adds it to the project.

  \sa removeForm()
*/

void QSProject::addForm( const QString &fileName )
{
    qas->addForm( fileName );
}

/*! Removes the form which was added under the filename \a fileName.

  \sa addForm()
*/

void QSProject::removeForm( const QString &fileName )
{
    qas->removeForm( fileName );
}

/*!
  \overload

  Reads and parses the Qt Script for Applications script from the \a
  projectData QByteArray. \a projectName must be specified.

  If an error occurs while running or parsing the script, the error()
  signal is emitted, and further action may occur depending on the
  errorMode().

  After open() is called, functions, classes and forms defined in this
  project can be called and used inside call() and evaluate() calls.

  If the project has signal/slot connections (for example, a form that
  connects widget signals to Qt Script slots), then these connections
  are established as part of the open() process.

  The project always stops when the QSA Developer is invoked, and
  continues when the QSA Developer is closed.
*/

void QSProject::open( const QByteArray &projectData, const QString &projectName )
{
    qas->open( projectData, projectName );
}

/*!
  Opens and parses the Qt Script for Applications script contained in
  the file, \a scriptProject. This overload of open() takes care of
  saving the project to the specified filename automatically. If an
  error occurs while running or parsing the script, actions according
  to the errorMode() are taken and the signal error() is emitted.

  After open() is called, functions, classes and forms defined in this
  project can be called and used using evaluate() and call().

  If the project has signal/slot connections (for example, a form that
  connects widget signals to Qt Script slots), then these connections
  are established as part of the open() process.

  The project always stops when QSA Developer is opened, and continues
  when QSA Developer is closed.
*/

void QSProject::open( const QString &scriptProject )
{
    qas->open( scriptProject );
}

/*! \enum QSProject::DeveloperFlags

  The DeveloperFlags describe in which mode QSA Developer is opened.

  \value GUIBuilder Opens QSA Developer with the GUI Builder

  \value Debugger Opens QSA Developer with the Debugger

  \value SignalHandlerTab Opens QSA Developer with the possibility to
  connect C++ signals to Qt Script function via the user interface

  \value EditableEditor Opens QSA Developer with editable editors

  \value FullDeveloper Convenience value to open QSA Developer with
  all features

  \value OnlyEditor Convenience value to open QSA Developer only with
  editor (including the signal handlers tab)

  \value OnlySimpleEditor Convenience value to open QSA Developer only with
  a simple editor and nothing else

  \value OnlyEditorAndDebugger Convenience value to open QSA Developer
  only with editor (including the signal handlers tab) and debugger
*/

/*!
  Opens QSA Developer with the current scripting project (if there is
  one open), so that the project can be edited and debugged.

  \a flags is a value which can contains values of DeveloperFlags or'd
  together.

  If there is no project currently open this function does nothing.

  This function clears the interpreter, so code which has been passed
  into previous evaluate() calls will be lost.

  If you created QSEditors, the QSA Developer is always opened with
  read only editors, even if EditableEditor is part of \a flags, so
  that the QSA Developer can be used for debugging, but not editing.
*/

void QSProject::openDeveloper( int flags )
{
    qas->openDeveloper( flags );
}

/*!
  Closes QSA Developer, if it is currently open.

  If there is no project currently open this function does nothing.

  This function clears the interpreter, so code which has been passed
  into previous evaluate() calls will be lost.
*/

void QSProject::closeDeveloper()
{
    qas->closeDeveloper();
}

/*!
  A function called \a functionName is added to the current project.
  The name must match the regular expression
  \c{/^[A-Za-z_][A-Za-z_0-9]*$/}.

  The code for the function can be passed in the \a functionBody
  parameter. If \a openDeveloper is TRUE (the default), then QSA
  Developer will be invoked so that the user can edit this function.
  If \a functionBody was specified its text will appear in the editor;
  if the \a functionBody is empty a new empty function will be
  created. If the function \a functionName already exists in the
  project and \a openDeveloper is TRUE, QSA Developer will be
  invoked showing the already existing function in the editor (and any
  \a functionBody will be discarded).

  This function returns TRUE on success (i.e. the \a functionName was
  valid); otherwise returns FALSE.

  If there is no project currently open this function does nothing.
*/

bool QSProject::addFunction( const QString &functionName,
			     const QString &functionBody,
			     bool openDeveloper )
{
    return qas->addFunction( functionName, functionBody, openDeveloper );
}

/*!
  Normally this function doesn't need to be called, since it is called
  automatically by the relevant functions (e.g. open(),
  call(), execute(), etc.). This function is useful in the case where
  an error has occurred and the script must be run again to
  reestablish the signal/slot connections between Qt and the script.
*/

void QSProject::reset()
{
    qas->reset();
}

